"""Flag provides a replacement for 'bool' which can't be instantiated or
managed correctly or something, so this is used by 'AttriBox' instances in
replacement of 'bool'."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

from worktoy.meta import BaseMetaclass


class _FlagMeta(BaseMetaclass):
  """Flag metaclass"""

  def __str__(cls, ) -> str:
    """String representation"""
    return type.__str__(bool)

  def __repr__(cls, ) -> str:
    """Code representation"""
    return type.__repr__(bool)

  def __instancecheck__(cls, instance) -> bool:
    """Instance check"""
    return bool.__instancecheck__(instance)

  def __subclasscheck__(cls, subclass) -> bool:
    """Subclass check"""
    return bool.__subclasscheck__(subclass)

  def __getattr__(cls, key: str) -> Any:
    """This method was implemented by a professional in a safe
    environment, do not try this at home!"""
    return type.__getattribute__(bool, key)


class Flag(metaclass=_FlagMeta):
  """Flag provides a replacement for 'bool' which can't be instantiated or
  managed correctly or something, so this is used by 'AttriBox' instances in
  replacement of 'bool'."""

  __inner_value__ = None

  def __init__(self, *args) -> None:
    if args:
      if args[0]:
        self.__inner_value__ = True
    else:
      self.__inner_value__ = False

  def __bool__(self) -> bool:
    """Bool representation"""
    return True if self.__inner_value__ else False

  def __getattribute__(self, key: str, **kwargs) -> Any:
    """This method was implemented by a professional in a safe
    environment, do not try this at home!"""
    if kwargs.get('_recursion', False):
      raise kwargs['_recursion']
    if key == '__class__':
      return bool
    if key == '__name__':
      return 'bool'
    if key == '__bases__':
      return (int,)
    if key == '__mro__':
      return bool, int, object
    if key == '__doc__':
      return bool.__doc__
    return object.__getattribute__(self, key)

  def __getattr__(self, key: str) -> Any:
    """This method was implemented by a professional in a safe
    environment, do not try this at home!"""
    try:
      if object.__getattribute__(self, '__inner_value__'):
        return object.__getattribute__(True, key)
      return object.__getattribute__(False, key)
    except AttributeError as attributeError:
      return type(self).__getattribute__(self, _recursion=attributeError)

  def __str__(self) -> str:
    """String representation"""
    if self.__inner_value__:
      return bool.__str__(True)
    return bool.__str__(False)

  def __repr__(self, ) -> str:
    """Code representation"""
    if self.__inner_value__:
      return bool.__repr__(True)
    return bool.__repr__(False)



"""BOX is a Zeroton meant to be used by the AttriBox class as a
placeholder for the box instance itself. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import Zeroton


class BOX(Zeroton):
  """BOX is a Zeroton meant to be used by the AttriBox class as a
  placeholder for the box instance itself. """



"""The 'worktoy.ezdata' class provides the EZData class!"""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._ez_space import EZSpace
from ._ez_meta import EZMeta
from ._ez_data import EZData



"""EZData provides a dataclass implementation. Subclasses of EZData must
place instances of AttriBox for each desired field. The order by which
they appear in the class body denote the order by which initial values
should be passed to the constructor. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.ezdata import EZMeta


class EZData(metaclass=EZMeta):
  """EZData provides a dataclass implementation. Subclasses of EZData must
  place instances of AttriBox for each desired field. The order by which
  they appear in the class body denote the order by which initial values
  should be passed to the constructor. """

  def __init__(self, *args, **kwargs) -> None:
    """This method is replaced during class creation. """



"""EZSpace provides the namespace object class for the EZData class."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import Callable
except ImportError:
  Callable = object

from worktoy.desc import AttriBox
from worktoy.meta import BaseNamespace
from worktoy.parse import maybe


class EZSpace(BaseNamespace):
  """EZSpace provides the namespace object class for the EZData class."""

  __field_boxes__ = None

  def _getFieldBoxes(self) -> list[tuple[str, AttriBox]]:
    """This method returns the field boxes."""
    return maybe(self.__field_boxes__, [])

  def _addFieldBox(self, key: str, box: AttriBox) -> None:
    """This method adds a field box to the namespace."""
    boxes = self._getFieldBoxes()
    self.__field_boxes__ = [*boxes, (key, box)]

  def __setitem__(self, key: str, value: object) -> None:
    """This method sets the key, value pair in the namespace."""
    if self.getClassName() != 'EZData':
      if key == '__init__':
        e = """EZData subclasses are not permitted to implement the 
        '__init__' method!"""
        raise AttributeError(e)
    if isinstance(value, AttriBox):
      self._addFieldBox(key, value)
    BaseNamespace.__setitem__(self, key, value)

  @staticmethod
  def _initFactory(attriBoxes: list[tuple[str, AttriBox]]) -> Callable:
    """This factory creates the '__init__' method which automatically
    populates the AttriBox instances."""

    keys = [key for (key, _) in attriBoxes]

    def __init__(self, *args, **kwargs) -> None:
      """This automatically generated '__init__' method populates the
      AttriBox instances."""
      for (key, arg) in zip(keys, args):
        setattr(self, key, arg)

      for key in keys:
        if key in kwargs:
          setattr(self, key, kwargs[key])

    return __init__

  def compile(self) -> dict:
    """The namespace created by the BaseNamespace class is updated with
    the '__init__' function created by the factory function."""
    out = BaseNamespace.compile(self)
    boxes = self._getFieldBoxes()
    if boxes:
      out['__init__'] = self._initFactory(boxes)
    return out



"""ATTR is a Zeroton meant to be used by the AttriBox class as a
placeholder for the box class itself. This would be AttriBox or a subclass
of it."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import Zeroton


class ATTR(Zeroton):
  """ATTR is a Zeroton meant to be used by the AttriBox class as a
  placeholder for the box class itself. This would be AttriBox or a subclass
  of it."""



"""BaseObject is used by AbstractMetaclass as a common baseclass for all
derived classes. All python classes are fundamentally subclasses of the
'object' class. However, two methods on the 'object' class, cause
problems. As of python 3.12.4, both '__init_subclass__' and '__init__'
will raise errors if invoked with arguments other than the subclass or
self respectively. While this is not without merit, flexible signatures
allowing arbitrary positional or keyword arguments are a common pattern.


class ParentClass:
  #  Base class example not implementing __init__ nor __init_subclass__.

class ChildClass(ParentClass):
  #  Child class which does implement __init__.

  def __init__(self, name: str=None, *args, **kwargs) -> None:
    self.name = 'unnamed' if name is None else name
    super().__init__(*args, **kwargs)  # This may raise an error

When instantiating the ChildClass, it will likely be working as expected.
But suppose a further subclass receives additional arguments:


class GrandChildClass(ChildClass):
  # Inspecting the code below would suggest that the __init__ method
  # is flexible enough to handle arbitrary positional arguments.

  def __init__(self, *args, **kwargs) -> None:
    super().__init__(*args, **kwargs)

Now consider:
someChild = GrandChildClass('Bob')  # No problem
anotherChild = GrandChildClass('Grace', 69)  # This will raise an error,
but why?

ChildClass('Grace', 69) leads to:
  anotherChild = GrandChildClass.__new__(GrandChildClass, 'Grace', 69)
  GrandChildClass.__init__(anotherChild, 'Grace', 69)
  #  The super call reaches the ChildClass:
  ChildClass.__init__(anotherChild, 'Grace', 69)
  #  Since ParentClass does not implement __init__, the super call instead
  #  invokes the __init__ on the first class in the method resolution order.
  #  Since ParentClass have no explicit baseclasses, this means 'object'.
  #  Thus:
  object.__init__(anotherChild, 69)
  #  But this leads to:
  #  'Exception: object.__init__() takes exactly one argument (the
  #  instance to initialize)'


A similar issues happen when applying keyword arguments (other than
metaclass=...) to the class creation, unless it is a baseclass of a class
that implements __init_subclass__. For example:

class SomeMeta(type):  # metaclass
  #  This metaclass tries to apply keyword arguments to the class after it
  #  is initialized.

  def __new__(mcls, name, bases, namespace, **kwargs) -> type:
    #  This method is included for illustration. The implementation is
    #  implements the same functionality as the default.
    return type.__new__(mcls, name, bases, namespace, **kwargs)

  def __init__(cls, name, bases, namespace, **kwargs) -> None:
    #  This method is called after the class object has been created.
    super().__init__(name, bases, namespace)
    for (key, val) in kwargs.items():
      setattr(cls, key, val)


class SomeClass(metaclass=SomeMeta, lmao=True):
  pass


In the previous example, the error message occurs when attempting to
instantiate the classes. But the above example will raise an error during
class creation itself:

"SomeClass.__init_subclass__() takes no keyword arguments"

To make use of keyword arguments during class creation, the functionality
must be implemented at the metaclass level. For example as in the above
example where the __init__ tries to apply the keyword arguments to the
newly created class. Even so, the above error still occurs. This error is
raised by the __new__ method in the metaclass when it calls the __new__
method on the 'type'. This means that in order to make use of keyword
arguments in class creation, not even a custom metaclass is sufficient.
The created class must also have as base class that implements
__init_subclass__ and stops the keyword arguments from reaching
object.__init_subclass__.

While there is merit in ensuring processing of arguments, this is
currently achieved at the expense of significant flexibility. For this
reason, the AbstractMetaclass will make use of the BaseObject class.
Please note, that the default implementations replaced are empty except
for raising the errors explained. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import BaseMetaclass


class BaseObject(metaclass=BaseMetaclass):
  """BaseObject provides argument-tolerant implementations of __init__ and
  __init_subclass__ preventing the errors explained in the documentation."""

  def __init__(self, *args, **kwargs) -> None:
    """Why are we still here?"""

  def __init_subclass__(cls, *args, **kwargs) -> None:
    """Just to suffer?"""



"""Dispatcher is the class responsible for calling the correct overloaded
function based on received arguments. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any, Optional

try:
  from typing import Callable
except ImportError:
  Callable = object
  
try:
  Overloaded = dict[tuple[type, ...], Callable]
except TypeError:
  Overloaded = dict


class Dispatcher:
  """Dispatcher is the class responsible for calling the correct overloaded
  function based on received arguments. """

  __overloaded_functions__ = None
  __bound_arg__ = None
  __static_method__ = None
  __class_method__ = None

  def __init__(self,
               overloadedFunctions: Overloaded,
               functionType: type) -> None:
    self.__overloaded_functions__ = overloadedFunctions
    if functionType is staticmethod:
      self.__static_method__ = True
      self.__class_method__ = False
    elif functionType is classmethod:
      self.__static_method__ = False
      self.__class_method__ = True
    else:
      self.__static_method__ = False
      self.__class_method__ = False

  def __call__(self, *args, **kwargs) -> Any:
    """The '__bound_arg__' provides the object to which this is bounded.
    If it covers a static method, no object is ever bounded, if a class
    method, the owner is bounded, and if an instance method, the instance is
    bounded. """
    if self.__bound_arg__ is None:  # Unbounded
      this = args[0]
      args = args[1:]
    else:  # bounded
      this = self.__bound_arg__
    typeSig = tuple(type(arg) for arg in args)
    func = self.__overloaded_functions__.get(typeSig, None)
    if func is None:
      floatSig = (*[float if arg is int else arg for arg in typeSig],)
      func = self.__overloaded_functions__.get(floatSig, None)
    if func is None:
      intArgs, intSig = [], []
      for arg in args:
        if isinstance(arg, float):
          if float.is_integer(arg):
            intArgs.append(arg)
            intSig.append(int)
          else:
            intArgs.append(arg)
            intSig.append(float)
        else:
          intArgs.append(arg)
          intSig.append(type(arg))
      intSig, intArgs = (*intSig,), (*intArgs,)
      func = self.__overloaded_functions__.get(intSig, None)
      if func is not None:
        return func(this, *intArgs, **kwargs)
    if func is None:
      for keyTypes, callMeMaybe in self.__overloaded_functions__.items():
        if len(keyTypes) == len(args):
          for keyType, arg in zip(keyTypes, args):
            if not isinstance(arg, keyType):
              break
          else:
            func = callMeMaybe
            break
      else:
        e = """Unable to match arguments to any overloaded function!"""
        raise ValueError(e)
    return func(this, *args, **kwargs)

  def __get__(self, instance: object, owner: type) -> Optional[Callable]:
    """Getter-function for descriptor protocol"""
    if self.__class_method__:
      self.__bound_arg__ = owner
    elif self.__static_method__:
      self.__bound_arg__ = None
    else:
      self.__bound_arg__ = instance
    return self



"""MissingValueException provides a subclass of Exception that subclasses
of AbstractDescriptor should raise to indicate a missing value. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

try:
  from typing import TYPE_CHECKING
except ImportError:
  TYPE_CHECKING = False

from worktoy.text import monoSpace, typeMsg

if TYPE_CHECKING:
  from worktoy.desc import AbstractDescriptor


class MissingValueException(Exception):
  """MissingValueException provides a subclass of Exception that subclasses
  of AbstractDescriptor should raise to indicate a missing value. """

  __instance_object__ = None
  __descriptor_object__ = None

  def __init__(self,
               descriptor: AbstractDescriptor) -> None:
    owner = descriptor.getFieldOwner()
    fieldName = descriptor.getFieldName()
    descType = type(descriptor).__name__
    msg = """The field '%s' of owning class '%s' managed by descriptor of 
    type '%s' received a call to '__get__' but no value
    was found!""" % (fieldName, owner.__name__, descType,)
    Exception.__init__(self, monoSpace(msg))

  def getInstanceObject(self, ) -> object:
    """Getter-function for the instance object. """
    return self.__instance_object__

  def getDescriptorObject(self, ) -> AbstractDescriptor:
    """Getter-function for the descriptor object. """
    if self.__descriptor_object__ is None:
      e = """The descriptor object has not been assigned!"""
      raise AttributeError(e)
    if isinstance(self.__descriptor_object__, AbstractDescriptor):
      return self.__descriptor_object__
    e = typeMsg('descriptorObject', self.__descriptor_object__,
                AbstractDescriptor)
    raise TypeError(e)

  def _setInstanceObject(self, instance: object) -> None:
    """Setter-function for the instance object. """
    self.__instance_object__ = instance

  def _setDescriptorObject(self, descriptor: AbstractDescriptor) -> None:
    """Setter-function for the descriptor object. """
    self.__descriptor_object__ = descriptor

  def __eq__(self, other: Any) -> bool:
    """The equality operator compares against a tuple to confirm that the
    exception involved the same instance and descriptor. """
    if self is other:
      return True
    if not isinstance(other, tuple):
      return NotImplemented
    if self.getInstanceObject() not in other:
      return False
    if self.getDescriptorObject() not in other:
      return False
    return True



"""The 'maybe' function receives any number of positional arguments and
returns the first argument that is not None. If all arguments are None,
the function returns None."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any


def maybe(*args: object) -> Any:
  """The 'maybe' function receives any number of positional arguments and
  returns the first argument that is not None. If all arguments are None,
  the function returns None."""
  for arg in args:
    if arg is not None:
      return arg
  return None



"""TYPE is a Zeroton serving as placeholder for the yet to be created
class owning the AttriBox instance."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import Callable
except ImportError:
  Callable = object

from worktoy.meta import Zeroton


class TYPE(Zeroton):
  """TYPE is a Zeroton serving as placeholder for the yet to be created
  class owning the AttriBox instance."""

  __call_me_maybe__ = None

  def __matmul__(self, callMeMaybe: Callable) -> None:
    self.__call_me_maybe__ = callMeMaybe

  def __call__(self, owner: type, *args, **kwargs) -> None:
    self.__call_me_maybe__(owner, *args, **kwargs)



"""The 'worktoy.parse' package provides tools for parsing data."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._maybe import maybe
from ._maybe_type import maybeType



"""This file provides some common type aliases used by 'worktoy.meta'. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import TypeAlias
except ImportError:
  TypeAlias = None

from typing import Union, Tuple

try:
  from typing import Callable
except ImportError:
  Callable = object
from worktoy.meta import AbstractNamespace


def functionInstance() -> None:
  pass


Function = type(functionInstance)

if TypeAlias is None:
  FunctionTypes = object
  Bases = object
  Namespace = object
  Space = object
  FunctionDecorator = object
  ClassDecorator = object
  Decorator = object
else:
  FunctionTypes: TypeAlias = Union[Function, classmethod, staticmethod]
  Bases: TypeAlias = Union[type, Tuple[type, ...]]
  Namespace: TypeAlias = Union[dict, AbstractNamespace]
  Space: TypeAlias = Namespace
  FunctionDecorator: TypeAlias = Callable[[Callable], Callable]
  ClassDecorator: TypeAlias = Callable[[type], type]
  Decorator: TypeAlias = Union[FunctionDecorator, ClassDecorator]



"""BaseNamespace provides the namespace object class for the
BaseMetaclass."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import Callable
except ImportError:
  Callable = object

from worktoy.meta import AbstractNamespace, Dispatcher
from worktoy.parse import maybe
from worktoy.text import typeMsg, monoSpace

try:
  Overloaded = dict[tuple[type, ...], Callable]
except TypeError:
  Overloaded = dict


class BaseNamespace(AbstractNamespace):
  """BaseNamespace provides the namespace object class for the
  BaseMetaclass."""

  __overloaded_functions__ = None

  def _getOverloadedFunctions(self) -> dict[str, Overloaded]:
    """Get the overloaded functions."""
    return maybe(self.__overloaded_functions__, {})

  def _setOverloadedFunctions(self, data: dict[str, Overloaded]) -> None:
    """Set the overloaded functions."""
    self.__overloaded_functions__ = data

  def _appendOverload(self, *args) -> None:
    """Append an overloaded function."""
    key, callMeMaybe, typeSig = [*args, None, None, None][:3]
    if key is None:
      e = """Unable to recognize key, function or type signature!"""
      raise ValueError(e)
    if not isinstance(key, str):
      e = typeMsg('key', key, str)
      raise TypeError(e)
    if callMeMaybe is None:
      e = """Unable to recognize function or type signature!"""
      raise ValueError(e)
    if not callable(callMeMaybe):
      e = typeMsg('callMeMaybe', callMeMaybe, Callable)
      raise TypeError(e)
    if typeSig is None:
      e = """Unable to recognize type signature!"""
      raise ValueError(e)
    self._validateTypeSignature(typeSig)
    allOverloads = self._getOverloadedFunctions()
    existing = allOverloads.get(key, {})
    if typeSig in existing:
      e = """Received duplicate signature '%s' for overloaded function!"""
      raise KeyError(monoSpace(e % typeSig))
    existing[typeSig] = callMeMaybe
    allOverloads[key] = existing
    self._setOverloadedFunctions(allOverloads)

  @classmethod
  def _validateTypeSignature(cls, *args) -> None:
    """This method validates the type signature. This method raises a type
    error if the type signature received fails to validate."""
    if len(args) == 1:
      if isinstance(args[0], (list, tuple)):
        return cls._validateTypeSignature(*args[0])
    for arg in args:
      if not isinstance(arg, type):
        e = typeMsg('arg', arg, type)
        raise TypeError(e)

  def __setitem__(self, key: str, value: object) -> None:
    """Set an item in the namespace."""
    if getattr(value, '__overloaded_signature__', None) is not None:
      typeSig = getattr(value, '__overloaded_signature__', )
      return self._appendOverload(key, value, typeSig)
    return AbstractNamespace.__setitem__(self, key, value)

  @staticmethod
  def _validateOverload(overloaded: Overloaded, key: str) -> type:
    """Validate the overloaded functions."""
    functionType = None
    for (typeSig, callMeMaybe) in overloaded.items():
      if functionType is None:
        functionType = type(callMeMaybe)
        continue
      if type(callMeMaybe) is not functionType:
        e = """Found overloaded functions at name: '%s' belonging to 
        different function types! All functions overloaded to the same 
        name must be of the same function type, either 'staticmethod', 
        'classmethod' or 'function'!"""
        raise TypeError(monoSpace(e % key))
    return functionType

  def compile(self) -> dict[str, object]:
    """Compile the namespace into a dictionary."""
    out = AbstractNamespace.compile(self, )
    for (key, overloaded) in self._getOverloadedFunctions().items():
      functionType = self._validateOverload(overloaded, key)
      out[key] = Dispatcher(overloaded, functionType)
    return out



"""THIS is a Zeroton meant to be used by the AttriBox class as a
placeholder for the instance of the class owning the AttriBox. This allows
the inner object managed by the AttriBox instance to include the owning
instance during class creation. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import Zeroton


class THIS(Zeroton):
  """THIS is a Zeroton serving as placeholder for the yet to be created
  instance owning the AttriBox instance. """



"""The overload function returns a callable that decorates a function with
the signature. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import Callable
except ImportError:
  Callable = object


def overload(*args, ):
  """The overload function returns a callable that decorates a function with
  the signature. """

  types = [a for a in args if isinstance(a, type)]

  def decorate(callMeMaybe: Callable) -> Callable:
    """Decorate the function with the given signature."""
    setattr(callMeMaybe, '__overloaded_signature__', (*types,))
    return callMeMaybe

  return decorate



"""Bag wraps the field object managed by an instance of AttriBox"""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from worktoy.meta import BaseObject
from worktoy.text import typeMsg


# noinspection PyMissingConstructor
class Bag(BaseObject):
  """Bag wraps the field object managed by an instance of AttriBox"""

  __owning_instance__ = None
  __owning_class__ = None
  __inner_object__ = None
  __inner_class__ = None

  def __init__(self, owningInstance: object, innerObject: object) -> None:
    self.setOwningInstance(owningInstance)
    self.setInnerObject(innerObject)

  def getOwningInstance(self) -> object:
    """Getter-function for the owning instance. """
    return self.__owning_instance__

  def setOwningInstance(self, owningInstance: object) -> None:
    """Setter-function for the owning instance. """
    if self.__owning_instance__ is not None:
      if self.__owning_instance__ is owningInstance:
        return
      e = """The owning instance has already been assigned!"""
      raise AttributeError(e)
    self.__owning_instance__ = owningInstance
    self.__owning_class__ = type(owningInstance)

  def getInnerObject(self) -> object:
    """Getter-function for the inner object. """
    return self.__inner_object__

  def setInnerObject(self, innerObject: object) -> None:
    """Setter-function for the inner object. """
    if self.__inner_class__ is None:
      self.__inner_object__ = innerObject
      self.__inner_class__ = type(innerObject)
    elif isinstance(innerObject, self.getInnerClass()):
      self.__inner_object__ = innerObject
    else:
      if self.getInnerClass() is complex:
        if isinstance(innerObject, (int, float)):
          return self.setInnerObject(complex(innerObject))
      if self.getInnerClass() is float and isinstance(innerObject, int):
        return self.setInnerObject(float(innerObject))
      if self.getInnerClass() is int and isinstance(innerObject, float):
        if innerObject.is_integer():
          return self.setInnerObject(int(innerObject))
      e = typeMsg('innerObject', innerObject, self.getInnerClass())
      raise TypeError(e)

  def getInnerClass(self) -> type:
    """Getter-function for the inner class. """
    return self.__inner_class__

  def getOwningClass(self) -> type:
    """Getter-function for the owning class. """
    return self.__owning_class__



"""AttriBox class improves the AttriBox class!"""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import Self
except ImportError:
  Self = object

try:
  from typing import Never
except ImportError:
  Never = object

from typing import Any

from worktoy.desc import Bag, THIS, TYPE, AbstractDescriptor, Flag, BOX, ATTR
from worktoy.parse import maybe
from worktoy.text import typeMsg, monoSpace


class AttriBox(AbstractDescriptor):
  """AttriBox class improves the AttriBox class!"""

  __field_class__ = None
  __pos_args__ = None
  __key_args__ = None

  @classmethod
  def __class_getitem__(cls, fieldClass: type) -> AttriBox:
    """Class method for creating a AttriBox instance."""
    return cls(fieldClass)

  def __init__(self, *args) -> None:
    AbstractDescriptor.__init__(self)
    for arg in args:
      if isinstance(arg, type):
        fieldClass = arg
        break
    else:
      e = """AttriBox constructor requires the fieldClass type!"""
      raise ValueError(e)
    if isinstance(fieldClass, type):
      self.__field_class__ = fieldClass
    else:
      e = """AttriBox constructor requires the fieldClass type!"""
      e2 = typeMsg('fieldClass', fieldClass, type)
      raise TypeError(monoSpace('%s\n%s' % (e, e2)))
    self.__field_class__ = fieldClass

  def __call__(self, *args, **kwargs) -> Self:
    self.__pos_args__ = [*args, ]
    self.__key_args__ = {**kwargs, }
    return self

  def getFieldClass(self, ) -> type:
    """Getter-function for the field class."""
    if self.__field_class__ is None:
      e = """The field class of the AttriBox instance has not been set!"""
      raise AttributeError(e)
    if self.__field_class__ is bool:
      return Flag
    if isinstance(self.__field_class__, type):
      return self.__field_class__
    e = typeMsg('__field_class__', self.__field_class__, type)
    raise TypeError(e)

  def getArgs(self, instance: object) -> list[Any]:
    """Getter-function for positional arguments"""
    args = []
    for arg in maybe(self.__pos_args__, []):
      if arg is THIS:
        args.append(instance)
      elif arg is TYPE:
        args.append(self.getFieldOwner())
      elif arg is BOX:
        args.append(self)
      elif arg is ATTR:
        args.append(type(self))
      else:
        args.append(arg)
    return args

  def getKwargs(self, instance: object) -> dict[str, Any]:
    """Getter-function for keyword arguments"""
    kwargs = {}
    for (key, val) in maybe(self.__key_args__, {}).items():
      if val is THIS:
        kwargs[key] = instance
      elif val is TYPE:
        kwargs[key] = self.getFieldOwner()
      else:
        kwargs[key] = val
    return kwargs

  def _createFieldObject(self, instance: object) -> Bag:
    """Create the field object."""
    if self.__field_class__ is None:
      e = """AttriBox instance has not been initialized!"""
      raise AttributeError(e)
    if self.__pos_args__ is None or self.__key_args__ is None:
      e = """AttriBox instance has not been called!"""
      raise AttributeError(e)
    args, kwargs = self.getArgs(instance), self.getKwargs(instance)
    fieldClass = self.getFieldClass()
    innerObject = fieldClass(*args, **kwargs)
    return Bag(instance, innerObject)

  def _getPrivateName(self, ) -> str:
    """Returns the name of the private attribute used to store the inner
    object. """
    if self.getFieldName() is None:
      e = """Instance of 'AttriBox' does not belong to class. This 
      typically indicates that the owning class is still being created."""
      raise RuntimeError(monoSpace(e))
    chars = []
    for (i, char) in enumerate(self.__field_name__):
      if char.isupper() and i and i < len(self.__field_name__):
        chars.append('_')
      chars.append(char.lower())
    innerName = ''.join(chars)
    innerNames = innerName.split('_')
    if len(innerNames) == 1:
      innerNames.append('value')
    innerName = '_'.join(innerNames)
    return """__%s__""" % (innerName,)

  def __instance_reset__(self, instance: object) -> None:
    """Reset-function for the instance."""
    pvtName = self._getPrivateName()
    delattr(instance, pvtName)

  def __instance_get__(self, instance: object, **kwargs) -> Any:
    """Getter-function for the instance."""
    pvtName = self._getPrivateName()
    if getattr(instance, pvtName, None) is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      setattr(instance, pvtName, self._createFieldObject(instance))
      return self.__instance_get__(instance, _recursion=True)
    bag = getattr(instance, pvtName)
    if not isinstance(bag, Bag):
      e = typeMsg('bag', bag, Bag)
      raise TypeError(e)
    innerObject = bag.getInnerObject()
    if innerObject is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      setattr(instance, pvtName, self._createFieldObject(instance))
      return self.__instance_get__(instance, _recursion=True)
    fieldClass = self.getFieldClass()
    if isinstance(innerObject, fieldClass):
      return innerObject
    e = typeMsg('innerObject', innerObject, fieldClass)
    raise TypeError(e)

  def __instance_set__(self, instance: object, value: object) -> None:
    """Setter-function for the instance."""
    pvtName = self._getPrivateName()
    bag = getattr(instance, pvtName, None)
    if bag is None:
      bag = self._createFieldObject(instance)
      bag.setInnerObject(value)
      return setattr(instance, pvtName, bag)
    if isinstance(bag, Bag):
      return bag.setInnerObject(value)
    e = typeMsg('bag', bag, Bag)
    raise TypeError(e)

  def __instance_del__(self, instance: object) -> Never:
    """Deleter-function for the instance."""
    e = """Deleter-function is not implemented by the AttriBox class."""
    raise TypeError(e)



"""AbstractMetaclass provides an abstract baseclass for custom
metaclasses. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import AbstractNamespace, Bases, Space


class MetaMetaclass(type):
  """MetaMetaclass is necessary to customize the __str__ method of a
  metaclass"""

  def __str__(cls) -> str:
    return cls.__name__


class AbstractMetaclass(MetaMetaclass, metaclass=MetaMetaclass):
  """The AbstractMetaclass class provides a base class for custom
  metaclasses."""

  @classmethod
  def __prepare__(mcls, name: str, bases: Bases, **kws) -> Space:
    """The __prepare__ method is invoked before the class is created. This
    implementation ensures that the created class has access to the safe
    __init__ and __init_subclass__ through the BaseObject class in its
    method resolution order."""
    return AbstractNamespace(mcls, name, bases, **kws)

  def __new__(mcls, name: str, bases: Bases, space: Space, **kws) -> type:
    """The __new__ method is invoked to create the class."""
    if hasattr(space, 'compile'):
      space = space.compile()
    return MetaMetaclass.__new__(mcls, name, bases, space, **kws)



"""Field provides a flexible implementation of the descriptor
protocol allowing owning classes to decorate methods as accessor methods. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import Callable
except ImportError:
  Callable = object

from worktoy.desc import AbstractDescriptor, MissingValueException
from worktoy.text import typeMsg


class Field(AbstractDescriptor):
  """Field provides a flexible implementation of the descriptor
  protocol allowing owning classes to decorate methods as accessor
  methods. """
  __field_type__ = None
  #  Accessor methods
  __getter_function__ = None
  __setter_function__ = None
  __resetter_function__ = None
  __deleter_function__ = None
  #  Keys for accessor methods
  __getter_key__ = None
  __setter_key__ = None
  __resetter_key__ = None
  __deleter_key__ = None

  def __set_name__(self, owner: type, name: str) -> None:
    """Set the name of the field and the owner of the field."""
    self.__field_owner__ = owner
    self.__field_name__ = name
    if self.__getter_key__ is not None:
      self.__getter_function__ = getattr(owner, self.__getter_key__, None)
    if self.__setter_key__ is not None:
      self.__setter_function__ = getattr(owner, self.__setter_key__, None)
    if self.__resetter_key__ is not None:
      self.__resetter_function__ = getattr(
          owner, self.__resetter_key__, None)
    if self.__deleter_key__ is not None:
      self.__deleter_function__ = getattr(owner, self.__deleter_key__, None)

  def getFieldType(self) -> type:
    """Getter-function for the field type."""
    if self.__field_type__ is None:
      return object
    if isinstance(self.__field_type__, type):
      return self.__field_type__
    e = typeMsg('__field_type__', self.__field_type__, type)
    raise TypeError(e)

  def __instance_reset__(self, instance: object) -> None:
    """Reset the instance object."""
    self.__get_resetter__()(instance)

  def __instance_get__(self, instance: object) -> object:
    """Get the instance object."""
    try:
      return self.__get_getter__()(instance)
    except AttributeError:
      raise MissingValueException(self)

  def __instance_set__(self, instance: object, value: object) -> None:
    """Set the instance object."""
    self.__get_setter__()(instance, value)

  def __instance_del__(self, instance: object) -> None:
    """Delete the instance object."""
    self.__get_deleter__()(instance)

  def __get_getter__(self, ) -> Callable:
    """Getter-function for the getter-function, getter-ception."""
    if self.__getter_function__ is None:
      raise AttributeError
    if callable(self.__getter_function__):
      return self.__getter_function__
    e = typeMsg('getter', self.__getter_function__, Callable)
    raise TypeError(e)

  def __get_setter__(self, ) -> Callable:
    """Getter-function for the setter-function of the field."""
    if callable(self.__setter_function__):
      return self.__setter_function__
    e = typeMsg('setter', self.__setter_function__, Callable)
    raise TypeError(e)

  def __get_resetter__(self, ) -> Callable:
    """Getter-function for the resetter-function of the field."""
    if callable(self.__resetter_function__):
      return self.__resetter_function__
    e = typeMsg('resetter', self.__resetter_function__, Callable)
    raise TypeError(e)

  def __get_deleter__(self, ) -> Callable:
    """Getter-function for the deleter-function of the field."""
    if callable(self.__deleter_function__):
      return self.__deleter_function__
    e = typeMsg('deleter', self.__deleter_function__, Callable)
    raise TypeError(e)

  def __set_getter__(self, callMeMaybe: Callable) -> Callable:
    """Set the getter function of the field."""
    if not callable(callMeMaybe):
      e = typeMsg('callMeMaybe', callMeMaybe, Callable)
      raise TypeError(e)
    self.__getter_key__ = callMeMaybe.__name__
    return callMeMaybe

  def __set_setter__(self, callMeMaybe: Callable) -> Callable:
    """Set the setter function of the field."""
    if not callable(callMeMaybe):
      e = typeMsg('callMeMaybe', callMeMaybe, Callable)
      raise TypeError(e)
    self.__setter_key__ = callMeMaybe.__name__
    return callMeMaybe

  def __set_resetter__(self, callMeMaybe: Callable) -> Callable:
    """Sets the resetter function of the field."""
    if not callable(callMeMaybe):
      e = typeMsg('callMeMaybe', callMeMaybe, Callable)
      raise TypeError(e)
    self.__resetter_key__ = callMeMaybe.__name__
    return callMeMaybe

  def __set_deleter__(self, callMeMaybe: Callable) -> Callable:
    """Set the deleter function of the field."""
    if not callable(callMeMaybe):
      e = typeMsg('callMeMaybe', callMeMaybe, Callable)
      raise TypeError(e)
    self.__deleter_key__ = callMeMaybe.__name__
    return callMeMaybe

  def GET(self, callMeMaybe: Callable) -> Callable:
    """Decorator for setting the getter function of the field."""
    return self.__set_getter__(callMeMaybe)

  def SET(self, callMeMaybe: Callable) -> Callable:
    """Decorator for setting the setter function of the field."""
    return self.__set_setter__(callMeMaybe)

  def RESET(self, callMeMaybe: Callable) -> Callable:
    """Decorator for setting the resetter function of the field."""
    return self.__set_resetter__(callMeMaybe)

  def DELETE(self, callMeMaybe: Callable) -> Callable:
    """Decorator for setting the deleter function of the field."""
    return self.__set_deleter__(callMeMaybe)



"""The EZMeta class provides the metaclass for the EZData class."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.ezdata import EZSpace
from worktoy.meta import BaseMetaclass, Bases


class EZMeta(BaseMetaclass):
  """The EZMeta class provides the metaclass for the EZData class."""

  @classmethod
  def __prepare__(mcls, name: str, bases: Bases, **kwargs) -> EZSpace:
    return EZSpace(mcls, name, bases, **kwargs)



"""AbstractDescriptor provides common boilerplate for descriptor classes. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from abc import abstractmethod
from typing import Any

from worktoy.desc import MissingValueException
from worktoy.meta import BaseMetaclass
from worktoy.parse import maybe
from worktoy.text import monoSpace, typeMsg


class AbstractDescriptor(metaclass=BaseMetaclass):
  """AbstractDescriptor provides common boilerplate for descriptor
  classes. """

  __field_owner__ = None
  __field_name__ = None
  __get_subscribers__ = None
  __set_subscribers__ = None
  __del_subscribers__ = None

  def __set_name__(self, owner: type, name: str) -> None:
    """Set the name of the field and the owner of the field."""
    self.__field_owner__ = owner
    self.__field_name__ = name

  def getFieldName(self) -> str:
    """Getter-function for the field name."""
    if self.__field_name__ is None:
      e = """Instance of 'AttriBox' does not belong to class. This 
      typically indicates that the owning class is still being created."""
      raise RuntimeError(monoSpace(e))
    if isinstance(self.__field_name__, str):
      return self.__field_name__
    e = typeMsg('__field_name__', self.__field_name__, str)
    raise TypeError(monoSpace(e))

  def getFieldOwner(self) -> type:
    """Getter-function for the field owner."""
    if self.__field_owner__ is None:
      e = """Instance of 'AttriBox' does not belong to class. This 
      typically indicates that the owning class is still being created."""
      raise RuntimeError(monoSpace(e))
    if isinstance(self.__field_owner__, type):
      return self.__field_owner__
    e = typeMsg('__field_owner__', self.__field_owner__, type)
    raise TypeError(monoSpace(e))

  def notifyGet(self, instance: object, value: object) -> None:
    for callMeMaybe in self._getGetSubscribers():
      callMeMaybe(instance, value)

  def notifySet(self,
                instance: object,
                oldValue: object,
                newValue: object) -> None:
    for callMeMaybe in self._getSetSubscribers():
      callMeMaybe(instance, oldValue, newValue)

  def notifyDel(self, instance: object, value: object) -> None:
    for callMeMaybe in self._getDelSubscribers():
      callMeMaybe(instance, value)

  def _getGetSubscribers(self) -> list[callable]:
    return maybe(self.__get_subscribers__, [])

  def _getSetSubscribers(self) -> list[callable]:
    return maybe(self.__set_subscribers__, [])

  def _getDelSubscribers(self) -> list[callable]:
    return maybe(self.__del_subscribers__, [])

  def _addGetSubscriber(self, callMeMaybe: callable) -> None:
    """Subscribes the callable received to be notified when the field
    getter is called. The instance and the value are passed as arguments.
    If instance is None, the AttriBox is accessed through the owning class,
    which does not result in a notification.

    Similar to this method, the _addSetSubscriber and _addDelSubscriber
    methods are used to subscribe callables to be notified when the field
    is accessed. """
    getSubscribers = self._getGetSubscribers()
    self.__get_subscribers__ = [*getSubscribers, callMeMaybe]

  def _addSetSubscriber(self, callMeMaybe: callable) -> None:
    """Registers the callable for notification when field setter is
    called."""
    setSubscribers = self._getSetSubscribers()
    self.__set_subscribers__ = [*setSubscribers, callMeMaybe]

  def _addDelSubscriber(self, callMeMaybe: callable) -> None:
    """Registers the callable for notification when field deleter is
    called."""
    delSubscribers = self._getDelSubscribers()
    self.__del_subscribers__ = [*delSubscribers, callMeMaybe]

  def ONGET(self, callMeMaybe: callable) -> callable:
    """Decorator for subscribing to the getter."""
    self._addGetSubscriber(callMeMaybe)
    return callMeMaybe

  def ONSET(self, callMeMaybe: callable) -> callable:
    """Decorator for subscribing to the setter."""
    self._addSetSubscriber(callMeMaybe)
    return callMeMaybe

  def ONDEL(self, callMeMaybe: callable) -> callable:
    """Decorator for subscribing to the deleter."""
    self._addDelSubscriber(callMeMaybe)
    return callMeMaybe

  def __get__(self, instance: object, owner: type) -> Any:
    """Get the value of the field."""
    if instance is None:
      return self
    try:
      value = self.__instance_get__(instance)
    except MissingValueException as missingValueException:
      raise AttributeError from missingValueException
    self.notifyGet(instance, value)
    return value

  def __set__(self, instance: object, value: object) -> None:
    """Set the value of the field."""
    try:
      oldValue = self.__instance_get__(instance)
    except MissingValueException:
      oldValue = None
    self.notifySet(instance, oldValue, value)
    self.__instance_set__(instance, value)

  def __delete__(self, instance: object) -> None:
    """Delete the value of the field."""
    try:
      oldValue = self.__instance_get__(instance)
    except MissingValueException as missingValueException:
      if missingValueException == (instance, self):
        oldValue = None
      else:
        raise missingValueException
    self.notifyDel(instance, oldValue)
    self.__instance_del__(instance)

  def __instance_reset__(self, instance: object) -> None:
    """Subclasses may implement this method to specify a way for the
    instance to have its value reset. """
    e = """The instance reset method is not implemented by the descriptor!"""
    raise TypeError(monoSpace(e))

  @abstractmethod
  def __instance_get__(self, instance: object) -> Any:
    """Subclasses should implement this method to provide the getter. """

  def __instance_set__(self, instance: object, value: object) -> None:
    """Subclasses should implement this method to provide the setter. """
    descName = self.__class__.__name__
    ownerName = self.getFieldOwner().__name__
    fieldName = self.getFieldName()
    e = """The attribute '%s' on class '%s' belongs to descriptor of type: 
    '%s' which does not implement setting!"""
    raise TypeError(monoSpace(e % (fieldName, ownerName, descName)))

  def __instance_del__(self, instance: object) -> None:
    """Subclasses should implement this method to provide the deleter. """
    descName = self.__class__.__name__
    ownerName = self.getFieldOwner().__name__
    fieldName = self.getFieldName()
    e = """The attribute '%s' on class '%s' belongs to descriptor of type: 
    '%s' which does not implement deletion!"""
    raise TypeError(monoSpace(e % (fieldName, ownerName, descName)))

  def reset(self, instance: object) -> None:
    """Reset the value of the field."""
    self.__instance_reset__(instance)
